import re
from eorg.const import TOKENS, METADATA


class Token:
    __slots__ = ["token", "value"]

    def __init__(self, token, value):
        self.token = token
        self.value = value

    def __repr__(self):
        return f"Token(token={self.token}, value={self.value})"


class Document:
    doc = []
    index = {}

    def __init__(self):
        self.doc = []
        self.index = {}

    def __getattr__(self, name, default=None):
        idx = self.index.get(name.upper(), [])
        if not idx:
            if default is not None:
                return default
            raise ValueError(f'Attribute of {name} does not exist in document')
        if len(idx) == 1:
            return self.doc[idx[0]].value
        return [self.doc[v].value for v in idx]

    def update(self, value):
        self.doc[-1].value += value

    def __iter__(self):
        for item in self.doc:
            yield item

    def body(self):
        for item in self.doc:
            if item.token in METADATA:
                continue
            yield item

    def __len__(self):
        return len(self.doc)

    def append(self, value):
        self.index.setdefault(value.token, []).append(len(self.doc))
        self.doc.append(value)


def parsebody(text, rx):
    match = re.search(rx, text)
    if match:
        return False, ""
    else:
        return rx, text + "\n"


def parseline(text):
    for key, (rx, block, s, e, count) in TOKENS.items():
        match = re.search(rx, text)
        if not match:
            continue
        level = len(match.group(0))
        if count is True:
            key += str(level)
        if key == "META":
            # return block, [match.group(0)[s:e], text[match.end():]]
            return (
                block,
                Token(token=match.group(0)[s:e], value=text[match.end() :]),
            )
        return block, Token(token=key, value=text[match.end() :])
    return False, Token(token='TEXT', value=text)


def parse(stream):
    doc = Document()
    block = False
    for line in stream:
        line = line.strip()
        if block is not False:
            result = parsebody(line, block)
            if result:
                block = result[0]
                doc.update(result[1])
                continue
        result = parseline(line)
        if result:
            block = result[0]
            doc.append(result[1])
    return doc
